import 'package:flutter/widgets.dart';
import 'package:flutter_extensions/flutter_extensions.dart';

class ChangeableValueBuilder<T extends Listenable, V> extends ChangeableConsumerBase<T> {
  final V Function(BuildContext context, T listenable) selector;
  final bool Function(V before, V after) condition;
  final Widget Function(BuildContext context, T listenable) builder;

  ChangeableValueBuilder({
    T listenable,
    @required this.selector,
    this.condition,
    @required this.builder,
  }) : super(listenable: listenable);

  @override
  _ChangeableValueBuilderState<T, V> createState() => _ChangeableValueBuilderState<T, V>();
}

class _ChangeableValueBuilderState<T extends Listenable, V>
    extends ChangeableConsumerBaseState<ChangeableValueBuilder<T, V>, T> {
  V _value;

  @override
  void updateListenable(T newListenable) {
    super.updateListenable(newListenable);
    _value = widget.selector(context, newListenable);
  }

  @override
  void updateState() {
    final newValue = widget.selector(context, listenable);
    if (_value != newValue || (widget.condition != null && widget.condition(_value, newValue))) {
      setState(() {
        _value = newValue;
      });
    }
  }

  @override
  Widget build(BuildContext context) {
    return widget.builder(context, listenable);
  }
}

class ChangeableBuilder<T extends Listenable> extends ChangeableConsumerBase<T> {
  final Widget Function(BuildContext context, T listenable) builder;

  ChangeableBuilder({T listenable, @required this.builder}) : super(listenable: listenable);

  @override
  _ChangeableBuilderState<T> createState() => _ChangeableBuilderState<T>();
}

class _ChangeableBuilderState<T extends Listenable>
    extends ChangeableConsumerBaseState<ChangeableBuilder<T>, T> {
  @override
  void updateState() {
    setState(() {});
  }

  @override
  Widget build(BuildContext context) {
    return widget.builder(context, listenable);
  }
}

abstract class ChangeableConsumerBase<T extends Listenable> extends StatefulWidget {
  final T listenable;

  const ChangeableConsumerBase({Key key, this.listenable}) : super(key: key);
}

abstract class ChangeableConsumerBaseState<W extends ChangeableConsumerBase<T>,
    T extends Listenable> extends State<W> {
  T _listenable;
  T get listenable => _listenable;

  @override
  void initState() {
    super.initState();
    updateListenable(widget.listenable);
  }

  @override
  void didUpdateWidget(W oldWidget) {
    super.didUpdateWidget(oldWidget);
    updateListenable(widget.listenable);
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    if (_listenable != null && widget.listenable != null) return;
    final newListenable = Provider.of<T>(context, listen: false);
    updateListenable(newListenable);
  }

  @override
  void dispose() {
    updateListenable(null);
    super.dispose();
  }

  void updateListenable(T newListenable) {
    if (_listenable == newListenable) return;
    _listenable?.removeListener(updateState);
    _listenable = newListenable;
    _listenable?.addListener(updateState);
  }

  void updateState();
}
